
/**
 * An implementation of Threefish encryption algorithm.
 *
 * Threefish is tweakable block encryption algorithm designed by Bruce Schneier,
 * Niels Ferguson, Stefan Lucks, Doug Whiting, Mihir Bellare, Tadayoshi Kohno,
 * Jon Callas, and Jesse Walker.
 *
 *
 * ISO C implementation by Alexander Myasnikov
 *
 * Web: www.darksoftware.narod.ru
 */



#include <mem.h>
#include <stdlib.h>

typedef unsigned long u32;
typedef unsigned long long u64;
typedef unsigned char u8;

#define blockSize 64
#define Nw 8
#define Nr 72

u32 p[8] = { 2, 1, 4, 7, 6, 5, 0, 3 };
u32 p_1[8] = { 6, 1, 0, 7, 2, 5, 4, 3 };

u32 r[8][4] =
  { {38, 30, 50, 53}, {48, 20, 43, 31}, {34, 14, 15, 27}, {26, 12, 58, 7},
{33, 49, 8, 42}, {39, 27, 41, 14}, {29, 26, 11, 9}, {33, 51, 39, 35}
};

u64 t[3];
u64 subKeys[19][8];



#define _mix(x, r, y)\
\
  y[0] = x[0] + x[1];\
  y[1] = (x[1] << r) | (x[1] >> (64 - r));\
  y[1] ^= y[0];\



#define _demix(y, r, x)\
\
  y[1] ^= y[0];\
  x[1] = (y[1] << (64 - (r))) | (y[1] >> (r));\
  x[0] = y[0] - x[1];\




void __stdcall __export
crypt (u64 * plain, u64 * c)
{
  u32 round;
  u64 f[8];
  u64 e[8];
  u64 y[2];
  u64 x[2];
  u64 v[8];
  u32 s;
  u32 i;

  memcpy (&v[0], plain, 64);


  for (round = 0; round < Nr; round++)
    {

      if (round % 4 == 0)
	{
	  s = round / 4;
	  for (i = 0; i < Nw; i++)
	    {
	      e[i] = v[i] + subKeys[s][i];
	    }
	}
      else
	{
	  for (i = 0; i < Nw; i++)
	    {
	      e[i] = v[i];
	    }
	}


      for (i = 0; i < Nw / 2; i++)
	{

	  x[0] = e[i * 2];
	  x[1] = e[i * 2 + 1];

	  _mix (x, r[round % 8][i], y);

	  f[i * 2] = y[0];
	  f[i * 2 + 1] = y[1];
	}

      for (i = 0; i < Nw; i++)
	{
	  v[i] = f[p[i]];
	}

    }

  for (i = 0; i < Nw; i++)
    {
      c[i] = v[i] + subKeys[Nr / 4][i];
    }

}




void __stdcall __export
decrypt (u64 * plain, u64 * c)
{

  u32 round;
  u64 f[8];
  u64 e[8];
  u64 y[2];
  u64 x[2];
  u64 v[8];
  u32 s;
  u32 i;

  memcpy (&v[0], plain, 64);

  for (round = Nr; round > 0; round--)
    {

      if (round % 4 == 0)
	{
	  s = round / 4;
	  for (i = 0; i < Nw; i++)
	    {
	      f[i] = v[i] - subKeys[s][i];
	    }
	}
      else
	{
	  for (i = 0; i < Nw; i++)
	    {
	      f[i] = v[i];
	    }
	}


      for (i = 0; i < Nw; i++)
	{
	  e[i] = f[p_1[i]];
	}

      for (i = 0; i < Nw / 2; i++)
	{

	  y[0] = e[i * 2];
	  y[1] = e[i * 2 + 1];

	  _demix (y, r[(round - 1) % 8][i], x);

	  v[i * 2] = x[0];
	  v[i * 2 + 1] = x[1];
	}

    }

  for (i = 0; i < Nw; i++)
    {
      c[i] = v[i] - subKeys[0][i];
    }

}


void __stdcall __export
setup (u64 * keyData, u64 * tweakData)
{
  u32 i, round;
  u64 K[8];
  u64 T[2];
  u64 key[9];

  u64 kNw = 6148914691236517205L;

  memcpy (&K[0], &keyData[0], 64);
  memcpy (&T[0], &tweakData[0], 16);


  for (i = 0; i < Nw; i++)
    {
      kNw ^= K[i];
      key[i] = K[i];
    }

  key[8] = kNw;

  t[0] = T[0];
  t[1] = T[1];
  t[2] = T[0] ^ T[1];



  for (round = 0; round <= Nr / 4; round++)
    {
      for (i = 0; i < Nw; i++)
	{
	  subKeys[round][i] = key[(round + i) % (Nw + 1)];

	  if (i == Nw - 3)
	    {
	      subKeys[round][i] += t[round % 3];
	    }
	  else if (i == Nw - 2)
	    {
	      subKeys[round][i] += t[(round + 1) % 3];
	    }
	  else if (i == Nw - 1)
	    {
	      subKeys[round][i] += round;
	    }
	}
    }

}
